#include "window.hpp"

#include "protocols.hpp"
#include "ui_window.h"

#include <QAbstractItemView>
#include <QCheckBox>
#include <QGuiApplication>
#include <QLabel>
#include <QShortcut>
#include <fstream>
#include <sys/socket.h>
#include <unistd.h>
#include <wayland-client-protocol.h>
#include <wayland-client.h>

static pid_t pid_from_fd(int fd) {
    struct ucred ucred;
    socklen_t len = sizeof(struct ucred);
    if (getsockopt(fd, SOL_SOCKET, SO_PEERCRED, &ucred, &len) == -1) {
        perror("getsockopt failed");
        exit(1);
    }
    return ucred.pid;
}

static std::string process_name_from_pid(const pid_t pid) {
    std::string procpath = QString::asprintf("/proc/%d/comm", pid).toStdString();

    std::ifstream infile(procpath);
    if (infile.is_open()) {
        std::string out;
        std::getline(infile, out);
        infile.close();
        return out;
    } else if (getenv("container") != nullptr) {
        // running in a flatpak, most likely
        return "Unknown (Sandboxed)";
    } else {
        return "Unknown";
    }
}

static void registry_global(void* data, wl_registry* registry, uint32_t name, const char* interface, uint32_t version) {
    (void) registry;
    (void) name;

    if (std::string(interface).starts_with("wl_")) {
        return;
    }

    auto* window = static_cast<Window*>(data);
    window->newGlobal(interface, version);
}

static void registry_global_remove(void* data, struct wl_registry* registry, uint32_t name) {
    (void) data;
    (void) registry;
    (void) name;
}

Window::Window(QNativeInterface::QWaylandApplication* waylandApp, QWidget* parent)
    : QMainWindow(parent), ui(std::make_unique<Ui::Window>()), searchBox(ui->toolbar), supportBox(ui->toolbar) {
    ui->setupUi(this);

    initTable(UPSTREAM, *ui->upstreamTable);
    initTable(WLROOTS, *ui->wlrootsTable);
    initTable(KDE, *ui->kdeTable);
    initTable(WESTON, *ui->westonTable);
    initTable(CHROMEOS, *ui->chromeosTable);
    initTable(UNKNOWN, *ui->unknownTable);

    for (auto protocol : known_protocols) {
        addProtocol(protocol.second);
    }

    ui->upstreamTable->sortByColumn(1, Qt::AscendingOrder);
    for (auto* table : {ui->upstreamTable, ui->wlrootsTable, ui->kdeTable, ui->westonTable, ui->chromeosTable}) {
        table->sortByColumn(0, Qt::AscendingOrder);
    }

    auto* display = waylandApp->display();
    auto fd = wl_display_get_fd(display);
    auto pid = pid_from_fd(fd);
    auto pname = process_name_from_pid(pid);

    auto listener = wl_registry_listener{.global = registry_global, .global_remove = registry_global_remove};

    auto* registry = wl_display_get_registry(display);
    wl_registry_add_listener(registry, &listener, (void*) this);
    wl_display_roundtrip(display);

    // Toolbar components
    QLabel* compositor = new QLabel(ui->toolbar);
    compositor->setText(QString::asprintf("Compositor: %s", pname.c_str()));
    compositor->setContentsMargins(4, 0, 0, 0);
    QFont font = compositor->font();
    font.setPointSize(14);
    compositor->setFont(font);
    ui->toolbar->addWidget(compositor);

    QWidget* spacer = new QWidget(ui->toolbar);
    spacer->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);
    ui->toolbar->addWidget(spacer);

    searchBox.setPlaceholderText("Search…");
    searchBox.setContentsMargins(4, 0, 4, 0);
    searchBox.setFixedWidth(16 * 14);
    QShortcut* shortcut = new QShortcut(QKeySequence(Qt::CTRL | Qt::Key_F), &searchBox);
    QObject::connect(shortcut, &QShortcut::activated, [this]() { searchBox.setFocus(); });

    ui->toolbar->addWidget(&searchBox);

    supportBox.addItem("All protocols", Model::supportFilters::ALL);
    supportBox.addItem("Supported protocols", Model::supportFilters::SUPPORTED);
    supportBox.addItem("Unsupported protocols", Model::supportFilters::UNSUPPORTED);
    ui->toolbar->addWidget(&supportBox);
}

Window::~Window() = default;

void Window::newGlobal(const std::string interface, const uint32_t version) {
    const Protocol& protocol = known_protocols[interface];
    if (protocol.source == UNKNOWN) {
        auto versionItem = new QStandardItem(false);
        versionItem->setCheckState(Qt::Checked);
        versionItem->setText(QString::asprintf("%d", version));

        auto items = QList<QStandardItem*>({new QStandardItem(QString::fromStdString(interface)), versionItem});

        for (auto item : std::as_const(items)) {
            item->setData(true, Qt::UserRole);
            auto font = item->font();
            font.setWeight(QFont::DemiBold);
            item->setFont(font);
            item->setData(true, Qt::UserRole);
        }

        models[protocol.source]->sourceModel.appendRow(items);
        return;
    }

    auto& model = models[protocol.source];

    auto column = (model->includeStatus) ? 2 : 1;
    auto matches = model->sourceModel.findItems(QString::fromStdString(protocol.name), Qt::MatchExactly, column);

    for (auto match : matches) {
        auto versionItem = model->sourceModel.item(match->row(), column + 1);
        versionItem->setCheckState(Qt::Checked);
        versionItem->setText(QString::asprintf("%d", version));

        for (auto i = 0; i < model->sourceModel.columnCount(); i++) {
            auto item = model->sourceModel.item(match->row(), i);
            auto font = item->font();
            font.setWeight(QFont::DemiBold);
            item->setFont(font);
            item->setData(true, Qt::UserRole);
        }
    }
}

void Window::addProtocol(const Protocol& protocol) {
    auto& model = models[protocol.source];

    auto versionItem = new QStandardItem(false);
    versionItem->setCheckState(Qt::Unchecked);
    versionItem->setText("N/A");

    auto items = QList<QStandardItem*>({
        new QStandardItem(QString::fromStdString(protocol.pretty)),
        new QStandardItem(QString::fromStdString(protocol.name)),
        versionItem,
    });
    if (protocol.source == UPSTREAM) {
        auto* statusItem = new QStandardItem(QString::fromStdString(status_to_string(protocol.status)));
        statusItem->setTextAlignment(Qt::AlignCenter);
        items.prepend(statusItem);
    }

    for (auto item : std::as_const(items)) {
        item->setData(false, Qt::UserRole);
    }

    model->sourceModel.appendRow(items);
}

void Window::initTable(const ProtocolSource source, QTableView& table) {
    models[source] = std::make_unique<Model>(source);
    auto& model = models[source];

    table.setModel(model->filteredModel);
    table.setSortingEnabled(true);
    table.setEditTriggers(QAbstractItemView::NoEditTriggers);
    table.setSelectionMode(QAbstractItemView::SingleSelection);
    table.verticalHeader()->setVisible(false);
    table.setItemDelegate(new TableItemDelegate);

    if (source == UPSTREAM) {
        table.horizontalHeader()->setSectionResizeMode(0, QHeaderView::ResizeToContents);
        table.horizontalHeader()->setSectionResizeMode(1, QHeaderView::Stretch);
        table.horizontalHeader()->setSectionResizeMode(2, QHeaderView::ResizeToContents);
        table.horizontalHeader()->setSectionResizeMode(3, QHeaderView::ResizeToContents);
    } else if (source != UNKNOWN) {
        table.horizontalHeader()->setSectionResizeMode(0, QHeaderView::ResizeToContents);
        table.horizontalHeader()->setSectionResizeMode(1, QHeaderView::Stretch);
        table.horizontalHeader()->setSectionResizeMode(2, QHeaderView::ResizeToContents);
    } else {
        table.horizontalHeader()->setSectionResizeMode(0, QHeaderView::Stretch);
        table.horizontalHeader()->setSectionResizeMode(1, QHeaderView::ResizeToContents);
    }

    table.sortByColumn(0, Qt::AscendingOrder);
    table.setEditTriggers(QAbstractItemView::NoEditTriggers);
    table.setSelectionMode(QAbstractItemView::SingleSelection);
    table.verticalHeader()->setVisible(false);
    table.setItemDelegate(new TableItemDelegate);
    table.setVerticalScrollBarPolicy(Qt::ScrollBarAlwaysOn);

    // Connect filters to UI
    QObject::connect(&searchBox, &QLineEdit::textChanged, [&model](const QString& text) { model->setSearchFilter(text); });

    QObject::connect(&supportBox, &QComboBox::currentIndexChanged, [&model, this](const int& index) {
        model->setSupportFilter(static_cast<Model::supportFilters>(supportBox.itemData(index).toInt()));
    });

    // Resize table columns when model is updated
    auto resizeColumns = [&table]() {
        table.setVisible(false);
        table.resizeColumnsToContents();
        table.setVisible(true);
    };

    QObject::connect(model->filteredModel, &QSortFilterProxyModel::rowsInserted, resizeColumns);
    QObject::connect(model->filteredModel, &QSortFilterProxyModel::rowsRemoved, resizeColumns);

    // Set tab title
    auto titleTab = [this, &model, &table](bool filtering) {
        int index = ui->toolBox->indexOf(&table);
        int count = model->filteredModel->rowCount();

        bool tabVisible = ui->toolBox->isTabVisible(index);
        bool tabShouldBeVisible = true;

        if (count == 0) {
            tabShouldBeVisible = false;
        } else if (!filtering) {
            ui->toolBox->setTabText(index, model->name);
        } else {
            ui->toolBox->setTabText(index, QStringLiteral("%1 (%2)").arg(model->name).arg(count));
        }

        if (tabVisible != tabShouldBeVisible) {
            ui->toolBox->setTabVisible(index, tabShouldBeVisible);
        }

        bool tabsVisible = false;
        for (int i = 0; i < ui->toolBox->count(); i++) {
            if (ui->toolBox->isTabVisible(i) == true) {
                tabsVisible = true;
                break;
            }
        }

        ui->toolBox->setVisible(tabsVisible);
        ui->filterMessage->setVisible(!tabsVisible);
    };

    titleTab(false);
    QObject::connect(model.get(), &Model::filterChanged, titleTab);
}

TableItemDelegate::TableItemDelegate(QObject* parent) : QStyledItemDelegate(parent) {}

void TableItemDelegate::paint(QPainter* painter, const QStyleOptionViewItem& option, const QModelIndex& index) const {
    bool enabled = index.data(Qt::UserRole).toBool();

    if (!enabled) {
        QStyleOptionViewItem newOption(option);
        initStyleOption(&newOption, index);

        QColor textColor = newOption.palette.text().color();
        textColor.setAlphaF(0.6);
        newOption.palette.setColor(QPalette::Text, textColor);

        QStyledItemDelegate::paint(painter, newOption, index);
    } else {
        QStyledItemDelegate::paint(painter, option, index);
    }
}
